/* Lite version, do not use in lessons! */
#include "PV112.h"

#define GLEW_STATIC
#if defined(_WIN32)
#define NOMINMAX      // Make Windows.h not define 'min' and 'max' macros
#include <GL/wglew.h> // Include on Windows
#else
#include <GL/glxew.h> // Include on Linux and Mac
#endif

#include <memory>
#include <sstream>
#include <fstream>
#include <iostream>

#include "cube.inl"

using namespace std;

namespace PV112
{

//------------------------------------------
//----    APPLICATION INITIALIZATION    ----
//------------------------------------------

void SetDebugCallback(GLDEBUGPROCARB callback)
{
    // Setup callback that will inform us when we make an error.
    // glDebugMessageCallbackARB is sometimes missed by glew, due to a bug in it.

#if defined(_WIN32)
    // On Windows, use this:
    PFNGLDEBUGMESSAGECALLBACKARBPROC myglDebugMessageCallbackARB =
        (PFNGLDEBUGMESSAGECALLBACKARBPROC)wglGetProcAddress("glDebugMessageCallbackARB");
    if (myglDebugMessageCallbackARB)
    {
        glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
        myglDebugMessageCallbackARB(callback, nullptr);
    }
#elif defined(__APPLE__)
    // On MacOS, use this (not tested):
    if (glDebugMessageCallbackARB)
    {
        glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
        glDebugMessageCallbackARB(callback, nullptr);
    }
#else
    // On Linux, use this:
    PFNGLDEBUGMESSAGECALLBACKARBPROC myglDebugMessageCallbackARB =
        (PFNGLDEBUGMESSAGECALLBACKARBPROC)glXGetProcAddress((unsigned char *)"glDebugMessageCallbackARB");
    if (myglDebugMessageCallbackARB)
    {
        glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
        myglDebugMessageCallbackARB(callback, nullptr);
    }
#endif
}

//------------------------------------
//----    SHADERS AND PROGRAMS    ----
//------------------------------------

string LoadFileToString(const char *file_name)
{
    ifstream file(file_name);
    stringstream ss;
    ss << file.rdbuf();
    return ss.str();
}

void WaitForEnterAndExit()
{
    cout << "Press Enter to exit" << endl;
    getchar();
    exit(1);
}

GLuint LoadAndCompileShader(GLenum shader_type, const char *file_name)
{
    // Load the file from the disk
    string s_source = LoadFileToString(file_name);
    if (s_source.empty())
    {
        cout << "File " << file_name << " is empty or failed to load" << endl;
        return 0;
    }

    // Create shader object and set the source
    GLuint shader = glCreateShader(shader_type);
    const char *source = s_source.c_str();
    glShaderSource(shader, 1, &source, nullptr);
    glCompileShader(shader);

    // Compile and get errors
    int compile_status;
    glGetShaderiv(shader, GL_COMPILE_STATUS, &compile_status);
    if (GL_FALSE == compile_status)
    {
        switch (shader_type)
        {
        case GL_VERTEX_SHADER:          cout << "Failed to compile vertex shader " << file_name << endl;                  break;
        case GL_FRAGMENT_SHADER:        cout << "Failed to compile fragment shader " << file_name << endl;                break;
        case GL_GEOMETRY_SHADER:        cout << "Failed to compile geometry shader " << file_name << endl;                break;
        case GL_TESS_CONTROL_SHADER:    cout << "Failed to compile tessellation control shader " << file_name << endl;    break;
        case GL_TESS_EVALUATION_SHADER: cout << "Failed to compile tessellation evaluation shader " << file_name << endl; break;
        default:                        cout << "Failed to compile shader " << file_name << endl;                         break;
        }

        int log_len = 0;
        glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &log_len);
        unique_ptr<char []> log(new char[log_len]);
        glGetShaderInfoLog(shader, log_len, nullptr, log.get());
        cout << log.get() << endl;

        glDeleteShader(shader);
        return 0;
    }
    else return shader;
}

GLuint CreateAndLinkProgram(const char *vertex_shader, const char *fragment_shader)
{
    // Load the vertex shader
    GLuint vs_shader = LoadAndCompileShader(GL_VERTEX_SHADER, vertex_shader);
    if (0 == vs_shader)
    {
        return 0;
    }

    // Load the fragment shader
    GLuint fs_shader = LoadAndCompileShader(GL_FRAGMENT_SHADER, fragment_shader);
    if (0 == fs_shader)
    {
        glDeleteShader(vs_shader);
        return 0;
    }

    // Create program and attach shaders
    GLuint program = glCreateProgram();
    glAttachShader(program, vs_shader);
    glAttachShader(program, fs_shader);
    glLinkProgram(program);

    // Link and get errors
    int link_status;
    glGetProgramiv(program, GL_LINK_STATUS, &link_status);
    if (GL_FALSE == link_status)
    {
        cout << "Failed to link program with vertex shader " << vertex_shader << " and fragment shader " << fragment_shader << endl;

        int log_len = 0;
        glGetProgramiv(program, GL_INFO_LOG_LENGTH, &log_len);
        unique_ptr<char []> log(new char[log_len]);
        glGetProgramInfoLog(program, log_len, nullptr, log.get());
        cout << log.get() << endl;

        glDeleteShader(vs_shader);
        glDeleteShader(fs_shader);
        glDeleteProgram(program);
        return 0;
    }
    else return program;
}

//-------------------------------------------
//----    SIMPLE PV112 GEOMETRY CLASS    ----
//-------------------------------------------

PV112Geometry::PV112Geometry()
{
    VertexBuffers[0] = 0;
    VertexBuffers[1] = 0;
    VertexBuffers[2] = 0;
    IndexBuffer = 0;
    VAO = 0;
    Mode = GL_POINTS;
    DrawArraysCount = 0;
    DrawElementsCount = 0;
}

PV112Geometry::PV112Geometry(const PV112Geometry &rhs)
{
    *this = rhs;
}

PV112Geometry &PV112Geometry::operator =(const PV112Geometry &rhs)
{
    VertexBuffers[0] = rhs.VertexBuffers[0];
    VertexBuffers[1] = rhs.VertexBuffers[1];
    VertexBuffers[2] = rhs.VertexBuffers[2];
    IndexBuffer = rhs.IndexBuffer;
    VAO = rhs.VAO;
    Mode = rhs.Mode;
    DrawArraysCount = rhs.DrawArraysCount;
    DrawElementsCount = rhs.DrawElementsCount;
    return *this;
}

void DeleteGeometry(PV112Geometry &geom)
{
    // This is mostly an example of what should be destroyed and how. We won't be using it anywhere.
    //
    // When using it, make sure the OpenGL context still exists (i.e. the main window still exists).

    // OpenGL silently ignores deleting objects that are 0, so this is safe even if the buffers were not created.
    glDeleteBuffers(3, geom.VertexBuffers);
    glDeleteBuffers(1, &geom.IndexBuffer);
    glDeleteVertexArrays(1, &geom.VAO);

    geom = PV112Geometry();        // Reset the state to 'no geometry'
}

void DrawGeometry(const PV112Geometry &geom)
{
    if (geom.DrawArraysCount > 0)
        glDrawArrays(geom.Mode, 0, geom.DrawArraysCount);
    if (geom.DrawElementsCount > 0)
        glDrawElements(geom.Mode, geom.DrawElementsCount, GL_UNSIGNED_INT, nullptr);
}

//-----------------------------
//----    BASIC OBJECTS    ----
//-----------------------------

PV112Geometry CreateCube(GLint position_location, GLint normal_location, GLint tex_coord_location)
{
    PV112Geometry geometry;

    // Create a single buffer for vertex data
    glGenBuffers(1, &geometry.VertexBuffers[0]);
    glBindBuffer(GL_ARRAY_BUFFER, geometry.VertexBuffers[0]);
    glBufferData(GL_ARRAY_BUFFER, cube_vertices_count * sizeof(float) * 8, cube_vertices, GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    geometry.VertexBuffers[1] = 0;
    geometry.VertexBuffers[2] = 0;

    // Create a buffer for indices
    glGenBuffers(1, &geometry.IndexBuffer);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, geometry.IndexBuffer);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, cube_indices_count * sizeof(unsigned int), cube_indices, GL_STATIC_DRAW);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

    // Create a vertex array object for the geometry
    glGenVertexArrays(1, &geometry.VAO);

    // Set the parameters of the geometry
    glBindVertexArray(geometry.VAO);
    glBindBuffer(GL_ARRAY_BUFFER, geometry.VertexBuffers[0]);
    if (position_location >= 0)
    {
        glEnableVertexAttribArray(position_location);
        glVertexAttribPointer(position_location, 3, GL_FLOAT, GL_FALSE, sizeof(float) * 8, 0);
    }
    if (normal_location >= 0)
    {
        glEnableVertexAttribArray(normal_location);
        glVertexAttribPointer(normal_location, 3, GL_FLOAT, GL_FALSE, sizeof(float) * 8, (const void *)(sizeof(float) * 3));
    }
    if (tex_coord_location >= 0)
    {
        glEnableVertexAttribArray(tex_coord_location);
        glVertexAttribPointer(tex_coord_location, 2, GL_FLOAT, GL_FALSE, sizeof(float) * 8, (const void *)(sizeof(float) * 6));
    }
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, geometry.IndexBuffer);

    glBindVertexArray(0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

    geometry.Mode = GL_TRIANGLES;
    geometry.DrawArraysCount = 0;
    geometry.DrawElementsCount = cube_indices_count;

    return geometry;
}

}
